import logging
import os
import sys
import argparse
import time
# handle pyserial module. It is not part of standard libraries
try:
    import serial
except ImportError:
    print("PySerialError: cannot find UART communication module 'PySerial'.")
    sys.exit(1)

import inspect
import itertools
import filecmp
import signal
from enum import Enum

sys.path.insert(0, "./")
from xymodem import XYmodem

DEFAULT_BL_BAUD = 460800
UART_TRANSMIT_RETRY_COUNT = 3

# directory to save the serial transaction logs.
LOG_FILE_PREFIX = 'dbg'
LOGS_DIR = 'logs'
WORKFILES_DIR = 'workFiles'
VERIFY_DIR = 'verify'

# to connect to the device
UART_INITIAL = b'COMM_START\n'
UART_INITIAL_ACK = b'COMM_START_ACK'
UART_TRANSMIT_READY = b'FLASH_MODE\n'
UART_TRANSMIT_READY_ACK = b'C'

UART_IMAGE_FLASH_ACK = b'DONE'
UART_IMAGE_FLASH_OK = b'FLASH_OK'
UART_IMAGE_FLASH_FAIL = b'FLASH_FAIL'
UART_IMAGE_YMOD_FAIL = b'YMOD_FAIL'

UART_NEXT_IMAGE = b'NEXT_IMAGE'

UART_READ_BACK_IMAGE = b'SNIMG\n'
UART_READ_BACK_DONE_ACK = b'UPLOADDONE'
SERIAL_RETRY = 3
SERIAL_M4_RETRY = 10

# to initiate device reboot after image flash is done (either success or failure, this handshake should be issued)
UART_REBOOT = b'DEV_REBOOT\n'
UART_REBOOT_ACK = b'DEV_REBOOT_ACK'

VERIFY_IMAGE_PREFIX = "VERIFY_"

#global lists holding the handshakes and expected ACK sequences.
image_flash_handshakes = [UART_INITIAL, UART_TRANSMIT_READY]
image_flash_handshake_acks = [UART_INITIAL_ACK, UART_TRANSMIT_READY_ACK]

image_flash_status = [UART_IMAGE_FLASH_OK, UART_IMAGE_FLASH_FAIL, UART_IMAGE_YMOD_FAIL]

reboot_confirm_handshake = [UART_REBOOT]
reboot_confirm_handshakes_ack = [UART_REBOOT_ACK]

next_image_handshake = [UART_NEXT_IMAGE]
next_image_handshake_ack = [UART_TRANSMIT_READY_ACK]

class FLASHTOOL(object):
    class flash_status(Enum):
        OK = 0
        FAIL = 1
        YM_FAIL = 2
        NONE = 3

    def __init__(self, flash_mode, comport, m4baud, files, verify=False, baud=DEFAULT_BL_BAUD):
        self.flash_mode = flash_mode
        self.comport = comport
        self.m4baud = m4baud
        self.files = files
        # enable this line if verify feature is visible in the GUI.
        #self.verify = verify
        self.verify = False
        self.baud = baud
        self.status = XYmodem.STATUS_FAIL
        self.retries = SERIAL_RETRY
        if(flash_mode):
            self.handshakes = image_flash_handshakes
            self.acks = image_flash_handshake_acks

    def wait_for_flash_ack(self, ser):
        read_data = bytearray([])
        got_ack = False
        ret_status = self.flash_status.NONE

        # send the handshake and check for the ACK signal in a loop.
        while 1:
            # get the time before read
            startTime = time.time()
            read_data.extend(ser.read(1))
            if (read_data.__contains__(UART_IMAGE_FLASH_OK)):
                logging.debug("%-24s %s",   "got ACK <<< ", read_data)
                ret_status = self.flash_status.OK
                got_ack = True
                break
            elif (read_data.__contains__(UART_IMAGE_FLASH_FAIL)):
                logging.debug("%-24s %s",   "got ACK <<< ", read_data)
                ret_status = self.flash_status.FAIL
                got_ack = True
                break
            elif (read_data.__contains__(UART_IMAGE_YMOD_FAIL)):
                logging.debug("%-24s %s",   "got ACK <<< ", read_data)
                ret_status = self.flash_status.YM_FAIL
                got_ack = True
                break

        # read current time and take the difference. Break the loop if it exceeds the timeout
            currentTime = time.time()
            if((currentTime - startTime) >= XYmodem.TIMEOUT):
                logging.debug("%-24s %s",   "bad ACK <<< ", read_data)
                break

        return got_ack, ret_status

    def wait_for_ack(self, ser, ack):
        read_data = bytearray([])
        got_ack = False
        # send the handshake and check for the ACK signal in a loop.
        while 1:
            # get the time before read
            startTime = time.time()
            read_data.extend(ser.read(1))
            if (read_data.__contains__(ack)):
                logging.debug("%-24s %s",   "got ACK <<< ", read_data)
                got_ack = True
                break
            # read current time and take the difference. Break the loop if it exceeds the timeout
            currentTime = time.time()
            if((currentTime - startTime) >= XYmodem.TIMEOUT):
                break

        return got_ack, read_data

    def send_signal_and_wait_for_ack(self, ser):
        for handshake, ack in zip(self.handshakes, self.acks):
            retry_count = 0
            read_data = bytearray([])
            # switch to BL baudrate before sending the READY command
            if (handshake == UART_TRANSMIT_READY):
                ser.baudrate = self.baud
                # flush out the UART input buffer
                ser.flushInput()

            while (True):
                ser.write(handshake)
                logging.debug("%-24s %s",   "sent signal >>> ", handshake)
                got_ack, read_data = self.wait_for_ack(ser, ack)
                if(got_ack == False):
                    retry_count += 1
                    if (retry_count >= self.retries):
                        logging.debug("Failed to get ACK")
                        return XYmodem.STATUS_FAIL
                    else:
                        logging.debug("%-24s %s",   "bad ACK <<< ", read_data)
                        continue
                else:
                    break;

        return XYmodem.STATUS_OK

    # initial handshake
    def establish_device_connection(self, ser):
        #1.assume the device is in the BL mode
        connect_status = self.send_signal_and_wait_for_ack(ser)
        #2.If not successful, try in the M4 APP mode (switch the baudrate to M4baudrate)
        if (connect_status ==XYmodem.STATUS_FAIL):
            logging.debug("%-24s %s",   "Device connection Failed in BL mode. ", "Trying in M4 APP mode.")
            ser.baudrate = self.m4baud # M4APP's baudrate
            self.retries = SERIAL_M4_RETRY # try more iterations than default retry
            # introduced a delay at BL to M4APP baudrate switch
            time.sleep(1)
            connect_status = self.send_signal_and_wait_for_ack(ser)
            self.retries = SERIAL_RETRY # reset the retries back to default value
        return connect_status

    # before sending next image, this handshake should be sent to tool
    def send_next_image_handshake(self, ser):
        self.handshakes = next_image_handshake
        self.acks = next_image_handshake_ack
        return self.send_signal_and_wait_for_ack(ser)

    # end of flash process, this handshake instruct the device to reboot.
    def send_reboot_confirm(self, ser):
        self.handshakes = reboot_confirm_handshake
        self.acks = reboot_confirm_handshakes_ack
        return self.send_signal_and_wait_for_ack(ser)

    def wait_for_flash_status_in_loop(self, ser):
        retry_count = 0
        read_data = bytearray([])

        while (True):
            got_ack, ret_status = self.wait_for_flash_ack(ser)
            if(got_ack == False):
                retry_count += 1
                if (retry_count >= self.retries):
                    logging.debug("Failed to get ACK")
                    return XYmodem.STATUS_FAIL, ret_status
                else:
                    continue
            else:
                break;

        return XYmodem.STATUS_OK, ret_status

    def wait_for_ack_in_loop(self, ser, ack):
        retry_count = 0
        read_data = bytearray([])

        while (True):
            got_ack, read_data = self.wait_for_ack(ser, ack)
            if(got_ack == False):
                retry_count += 1
                if (retry_count >= self.retries):
                    logging.debug("Failed to get ACK")
                    return XYmodem.STATUS_FAIL
                else:
                    logging.debug("%-24s %s",   "bad ACK <<< ", read_data)
                    continue
            else:
                break;

        return XYmodem.STATUS_OK

    def flash_image(self):
        #Open the serial Port
        ser = serial.Serial()
        ser.port = self.comport
        ser.baudrate = self.baud
        ser.timeout = XYmodem.TIMEOUT
        ser.write_timeout = XYmodem.TIMEOUT

        try:
            ser.open()
        except Exception as e:
            logging.debug("%-24s %s",   "SerialError", e)
            logging.error("%-24s %s. %s",   "Failed to open comport:", ser.name, "make sure the comport is available")
            return self.status

        logging.debug("%-24s %s",   "Serial comport successfully opened: ", ser.name)
        # try for device connection
        connect_status = self.establish_device_connection(ser)

        if (connect_status == XYmodem.STATUS_OK):
            logging.debug("%-24s",   "Device connection successful!")
            ser.baudrate = self.baud
            sender = XYmodem(XYmodem.PROTOCOL_YMODEM, ser, True)

            # flush out the UART input buffer
            ser.flushInput()

            # now start flashing.
            parentDir = os.path.normpath(os.getcwd()+os.sep+os.pardir)
            verify_dir = VERIFY_DIR + self.comport
            recv_dir = os.path.join(parentDir, WORKFILES_DIR, verify_dir)
            read_back_status = XYmodem.STATUS_OK

            # clean the folder if it already exist, otherwise create one
            if(self.verify):
                if os.path.exists(recv_dir):
                    filelist = os.listdir(recv_dir)
                    for f in filelist:
                        item = os.path.join(recv_dir,f)
                        try:
                            os.remove(item)
                        except OSError:
                            pass
                else:
                    os.makedirs(recv_dir)

            num_files = len(self.files)
            current_file_index = 1

            for image in self.files:
                # enable this line if verify feature is visible in the GUI.
                #time.sleep(5)

                ser.flushInput()

                self.status = sender.send([image])
                if (self.status == XYmodem.STATUS_OK):

                    fv_status = self.flash_status.NONE
                    ret, fv_status = self.wait_for_flash_status_in_loop(ser)

                    # reset final status here.
                    self.status = XYmodem.STATUS_FAIL

                    if (ret == XYmodem.STATUS_OK):
                        # something went wrong?
                        if (fv_status == self.flash_status.NONE):
                            logging.debug("%-24s %s",   "Unexpected device response while flashing image, ", image)
                            break
                        #flash or verify failed
                        elif (fv_status == self.flash_status.FAIL):
                            logging.debug("%-24s %s", "Failed to flash image ", image)
                            break
                        #ymodem receive failed in the device.
                        elif (fv_status == self.flash_status.YM_FAIL):
                            logging.debug("%-24s %s", "File transfer failed for image ", image)
                            break
                        # flash and verify are successful
                        elif (fv_status == self.flash_status.OK):
                            logging.debug("%-24s %s",   "Successfully flashed image :", image)
                            # send next-image handshake if current image is not the last file in the list
                            if (current_file_index < num_files):
                                current_file_index += 1
                                next_image_status = self.send_next_image_handshake(ser)
                                if (next_image_status == XYmodem.STATUS_FAIL):
                                    # report if any failure and break
                                    logging.debug("%-24s %s", "Next image handshake failed to get acknowledged after flashing image ", image)
                                    break
                            # set status to success and continue for next image flash
                            self.status = XYmodem.STATUS_OK
                            continue
                    else:
                        logging.debug("%-24s %s",   "Failed to receive flash status for image, ", image)
                        break

                    '''
                    # this code has to be enabled when verify feature is visible in the GUI.
                    # continue if verify flag is not set
                    if not self.verify:
                        continue
                    #now read-back the flashed image
                    read_back_status = self.wait_for_ack_in_loop(ser,UART_IMAGE_FLASH_ACK)
                    if (read_back_status == XYmodem.STATUS_OK):
                        # send handshake to initiate image read-back
                        ser.write(UART_READ_BACK_IMAGE)
                        logging.debug("%-24s %s",   "sent read-back ready signal >>> ", UART_READ_BACK_IMAGE)
                        # have a delay to avoid debug prints from BL.
                        time.sleep(5)
                        # read the image back from flash memory
                        read_back_status = sender.recv(recv_dir)
                        if (read_back_status == XYmodem.STATUS_OK):
                            # wait for image read-back ack from device.
                            time.sleep(5)
                            read_back_status = self.wait_for_ack_in_loop(ser, UART_READ_BACK_DONE_ACK)
                            if (read_back_status == XYmodem.STATUS_OK):
                                # compare the original image and read-back image
                                verify_image = os.path.join(recv_dir, (VERIFY_IMAGE_PREFIX + os.path.basename(image)))

                                # verify_status = filecmp.cmp(image, verify_image, shallow=False)
                                if (filecmp.cmp(image, verify_image, shallow=False) == True):
                                    logging.debug("%-24s %s",   "Successfully verified image, ", image)
                                    continue
                                else:
                                    # record the fail status and break
                                    read_back_status = XYmodem.STATUS_FAIL
                                    logging.debug("%-24s %s",   "Image verify failed for ", image)
                                    break
                            else:
                                logging.debug("%-24s %s",   "Verify failed as read-back got interrupted for image ", image)
                                break
                        else:
                            logging.debug("%-24s %s",   "Verify failed as read-back failed for ", image)
                            break
                    else:
                        logging.debug("%-24s %s", "Could not start read-back. Failed to get image receive ACK from device for ", image)
                        break
                    '''

                else:
                    logging.debug("%-24s %s", "Failed to flash image ", image)
                    break

            # enable this line if verify feature is visible in the GUI.
            #self.status |= read_back_status

            if (self.status == XYmodem.STATUS_OK):
                if (self.verify):
                    logging.info("%s: %s",self.comport, "Image(s) successfully flashed and verified!")
                    logging.debug("%-24s %s", "Selected image(s), ", self.files)
                else:
                    logging.info("%s: %s",self.comport, "Image(s) successfully flashed!")
                    logging.debug("%-24s %s", "Selected image(s), ", self.files)
            else:
                logging.error("%s: %s",self.comport, "Flashing failed.")
                logging.debug("%-24s %s", "Selected image(s)", self.files)

            # reboot the BL before returning
            reboot_status = self.send_reboot_confirm(ser)

            # log the final status into console/file and return the status.
            if (reboot_status == XYmodem.STATUS_OK):
                logging.debug("%-24s",   "Device reboot command sent and acknowledged")
            else:
                logging.debug("%-24s",   "Failed to Send Device Reboot Command.")

        else:
            logging.debug("%-24s",   "Device connection Failed.")
            logging.error("%s: %s",self.comport, "Flashing failed.")
            logging.debug("%-24s %s", "Selected image(s)", self.files)

        ser.close()
        return (self.status or reboot_status)

    # def get_ramdump(self):

def sig_handler(signum, frame):
    print("CTRL+C!")
    print(signum)
    print(frame)
    sys.exit()

def main(argv=None):
    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    # setup argument parser
    parser = argparse.ArgumentParser(description = "", formatter_class = argparse.RawDescriptionHelpFormatter)

    # flash? or ramdump?
    parser_group_fl_rdump = parser.add_mutually_exclusive_group(required=True)
    parser_group_fl_rdump.add_argument("--flash", action="store_true", help="")
    parser_group_fl_rdump.add_argument("--rdump", action="store_false", help="")

    #verify the image in the UART flash mode using a read-back from the device
    parser.add_argument("--verify", dest="verify", action="store_true", help="")

    # com port and baud ?
    parser.add_argument("--port", dest="port", help="", required=True)
    parser.add_argument("--baud", dest="baud", help="", type=int, required=True)

    # file(s) to send ?
    parser.add_argument('--files', dest='files', default=[], help="", nargs="+")

    # Process arguments
    args = parser.parse_args()

    fileName = LOG_FILE_PREFIX + args.port + '.log'
    parentDir = os.path.normpath(os.getcwd()+os.sep+os.pardir)
    logFileDir = os.path.join(parentDir, LOGS_DIR)

    if not os.path.exists(logFileDir):
        os.mkdir(logFileDir)

    logFile = os.path.join(logFileDir, fileName)

    if os.path.exists(logFile):
        try:
            os.remove(logFile)
        except OSError:
            pass

    # logging
    stream_handler = logging.StreamHandler(sys.stdout)
    #stream_handler_formatter = logging.Formatter("[%(filename)-18s:%(lineno)4d  %(funcName)24s()] %(message)s")
    stream_handler_formatter = logging.Formatter("%(message)s")
    stream_handler.setFormatter(stream_handler_formatter)
    stream_handler.setLevel(logging.INFO)

    file_handler = logging.FileHandler(logFile, "w")
    file_handler_formatter = logging.Formatter('[%(asctime)s %(filename)18s:%(lineno)4s %(funcName)24s()] %(levelname)-8s %(message)s')
    file_handler.setFormatter(file_handler_formatter)
    file_handler.setLevel(logging.DEBUG)

    logging.basicConfig(level=logging.DEBUG, handlers = [stream_handler, file_handler])

    signal.signal(signal.SIGINT, sig_handler)

    logging.debug("%-24s [%s]", "port", args.port)
    logging.debug("%-24s [%d]", "baud", args.baud)
    if (args.flash):
        logging.debug("%-24s [%s]", "tool", "device-flash")
    else:
        logging.debug("%-24s [%s]", "tool", "ram-dump")

    logging.debug("%-24s [%s]", "file(s)", args.files)

    tool = FLASHTOOL(args.flash, args.port, args.baud, args.files, args.verify)

    if (tool.flash_mode):
        return tool.flash_image()

if __name__ == "__main__":
    sys.exit(main())


